//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : FFTProcess
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "fft_process.h"

//============================================================================//
void FFTProcess::alloc_array(const int &num_fft)
//============================================================================//
{
    _Num_fft = num_fft;

    const int     nf3     = _Num_fft * _Num_fft * _Num_fft;
    const int     ns  = _parameter->_Num_sort;
    const int     nz  = _parameter->_NZangle;
    const int     no  = _parameter->_Num_output;
    const int nproc2  = _parallel->nproc2();
    int   nt;

    nt = nz * ns;
    if( no > nt ) nt = no;

    alloc_fft();

    _Select.resize(nproc2);
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _Select[i].resize(ns);
    }

    _Top.resize(nt);

    //---------- memory allocation for _Rotation
    _Rotation = new int[nproc2];

    //---------- memory allocation for _FFT_rec_r
    _FFT_rec_r = new float[nf3];
    if( !_FFT_rec_r ) {
        cerr << "[ERROR] Out of memory. Number of listed receptors = ("
             << nf3 << ") for (_FFT_rec_r) in fft_process.cpp!!\n";
        exit(1);
    }

    //---------- memory allocation for _FFT_rec_i
    _FFT_rec_i = new float[nf3];
    if( !_FFT_rec_i ) {
        cerr << "[ERROR] Out of memory. Number of listed receptors = ("
             << nf3 << ") for (_FFT_rec_i) in fft_process.cpp!!\n";
        exit(1);
    }

    //---------- memory allocation for _FFT_lig_r
    _FFT_lig_r = new float*[nproc2];
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _FFT_lig_r[i] = new float[nf3];
        if( !_FFT_lig_r[i] ) {
            cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
                 << nf3*nproc2 << ") for (_FFT_lig_r) in fft_process.cpp!!\n";
            exit(1);
        }
    }

    //---------- memory allocation for _FFT_lig_i
    _FFT_lig_i = new float*[nproc2];
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _FFT_lig_i[i] = new float[nf3];
        if( !_FFT_lig_i[i] ) {
            cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
                 << nf3*nproc2 << ") for (_FFT_lig_i) in fft_process.cpp!!\n";
            exit(1);
        }
    }

    return;
}

//============================================================================//
void FFTProcess::alloc_fft()
//============================================================================//
{
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int nproc2  = _parallel->nproc2();


    _FFTWin  = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex)*nf3*nproc2);
    _FFTWout = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex)*nf3*nproc2);

    return;
}

//============================================================================//
void FFTProcess::receptor_fft(float *grid_r,float *grid_i)
//============================================================================//
{
    const int num_grid= _Num_fft / 2;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int ndata   = ( _Num_fft - num_grid ) / 2;
    const float   theta   = -2.0 * PI / _Num_fft;

    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFTWin[i][0] = 0.0;
        _FFTWin[i][1] = 0.0;
    }

    for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
        const int ic = _Num_fft*_Num_fft*(i+ndata);

        for( int j = 0 ; j < num_grid ; j++ ) {
            const int jc = ic + _Num_fft*(j+ndata);

            for( int k = 0 ; k < num_grid ; k++ ) {
                _FFTWin[jc+k+ndata][0] = grid_r[m  ];
                _FFTWin[jc+k+ndata][1] = grid_i[m++];
            }
        }
    }

    fft3d(theta,0);

    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFT_rec_r[i] = _FFTWout[i][0];
        _FFT_rec_i[i] = _FFTWout[i][1];
    }

    return;
}

//============================================================================//
void FFTProcess::ligand_fft(float *grid_r,float *grid_i,int myid2)
//============================================================================//
{
    const int     num_grid = _Num_fft / 2;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int ndata   = ( _Num_fft - num_grid ) / 2;
    const float   theta   = -2.0 * PI / _Num_fft;

    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFTWin[nf3*myid2+i][0] = 0.0;
        _FFTWin[nf3*myid2+i][1] = 0.0;
    }

    for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
        const int ic = _Num_fft*_Num_fft*(i+ndata);

        for( int j = 0 ; j < num_grid ; j++ ) {
            int jc = ic + _Num_fft*(j+ndata);

            for( int k = 0 ; k < num_grid ; k++ ) {
                _FFTWin[nf3*myid2+jc+k+ndata][0] = grid_r[m  ];
                _FFTWin[nf3*myid2+jc+k+ndata][1] = grid_i[m++];
            }
        }
    }

    fft3d(theta,myid2);

    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFT_lig_r[myid2][i] = _FFTWout[nf3*myid2+i][0];
        _FFT_lig_i[myid2][i] = _FFTWout[nf3*myid2+i][1];
    }

    return;
}

//============================================================================//
void FFTProcess::fft_score(int myid2)
//============================================================================//
{
    const int nf3     = _Num_fft * _Num_fft * _Num_fft;
    const float   theta   = 2.0 * PI / _Num_fft;


    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFTWin[nf3*myid2+i][0] = _FFT_rec_r[i]*_FFT_lig_r[myid2][i]
                                  + _FFT_rec_i[i]*_FFT_lig_i[myid2][i];
        _FFTWin[nf3*myid2+i][1] = _FFT_rec_r[i]*_FFT_lig_i[myid2][i]
                                  - _FFT_rec_i[i]*_FFT_lig_r[myid2][i];
    }

    fft3d(theta,myid2);

    return;
}

//============================================================================//
void FFTProcess::fft3d(const float &theta,int myid2)
//============================================================================//
{
    const int n   = _Num_fft;
    const int nf3 = n*n*n;


    fftwf_plan    plan;

    #pragma omp critical
    {
        if( theta < 0.0 ) {
            plan=fftwf_plan_dft_3d(n,n,n,&_FFTWin[nf3*myid2],&_FFTWout[nf3*myid2],
                                   FFTW_FORWARD,FFTW_ESTIMATE);
        } else {
            plan=fftwf_plan_dft_3d(n,n,n,&_FFTWin[nf3*myid2],&_FFTWout[nf3*myid2],
                                   FFTW_BACKWARD,FFTW_ESTIMATE);
        }
    }

    fftwf_execute(plan);

    #pragma omp critical
    {
        fftwf_destroy_plan(plan);
    }


    return;
}

//============================================================================//
void FFTProcess::score_sort(int myid2)
//============================================================================//
{
    const int ns  = _parameter->_Num_sort;
    const int nf2 = _Num_fft * _Num_fft;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;

    for( int i = 0 ; i < ns ; i++ ) {
        _Select[myid2][i].score = 0.0;
    }

    if(ns!=1) {
        for( int i = 0 ; i < nf3 ; i++ ) {

            const float raw = _FFTWout[nf3*myid2+i][0] / nf3;

            for( int j = 0 ; j < ns ; j++ ) {
                if( raw > _Select[myid2][j].score ) {
                    for( int k = ns-1 ; k > j ; k-- ) {
                        _Select[myid2][k] = _Select[myid2][k-1];
                    }
                    _Select[myid2][j].score    = raw;
                    _Select[myid2][j].index[1] = i / nf2;
                    _Select[myid2][j].index[2] = (i / _Num_fft) % _Num_fft;
                    _Select[myid2][j].index[3] = i % _Num_fft;
                    break;
                }
            }
        }
    } else { // ns = 1, take only 1 score per angle
        float temp_top_score = 0.0;
        int temp_top_num = 0;
        for( int i = 0 ; i < nf3 ; i++ ) {

            const float raw = _FFTWout[nf3*myid2+i][0];

            if (temp_top_score < raw) {
                temp_top_score = raw;
                temp_top_num = i;
            }
        }
        _Select[myid2][0].score    = temp_top_score / nf3;
        _Select[myid2][0].index[1] = temp_top_num / nf2;
        _Select[myid2][0].index[2] = (temp_top_num / _Num_fft) % _Num_fft;
        _Select[myid2][0].index[3] = temp_top_num % _Num_fft;
    }

    for( int i = 0 ; i < ns ; i++ ) {
        _Select[myid2][i].index[0] = _Rotation[myid2];
    }

    for( int i = 0 ; i < ns ; i++ ) {
        _Top[_Rotation[myid2]*ns+i] = _Select[myid2][i];
    }

    return;
}

//============================================================================//
bool ascending_struct(const SortScore &s1,const SortScore &s2)  // for sort
//============================================================================//
{
    return s1.score > s2.score;
}

//============================================================================//
void FFTProcess::sort_index(float *fwork,int *iwork)
//============================================================================//
{
    const int     no  = _parameter->_Num_output;
    const int     nt  = (_Top.size() < no) ? _Top.size() : no;

    partial_sort(_Top.begin(),_Top.begin()+nt,_Top.end(),
                 ascending_struct);

    return;
}

//============================================================================//
void FFTProcess::top_score_clean()
//============================================================================//
{
    const int     ns  = _parameter->_Num_sort;
    const int     nz  = _parameter->_NZangle;
    const int     no  = _parameter->_Num_output;
    int   nt;

    nt = nz * ns;
    if( no > nt ) nt = no;

    for( int j = 0 ; j < nt ; j++ ) {
        _Top[j].score = 0.0;
    }

    return;
}
