//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : (main)
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "cpu_time.h"
#include "control.h"

#define VERSION "2.6"

#define LASTUPDATED "2013/01/11"

//============================================================================//
int main(int argc,char *argv[])
//============================================================================//
{
    Parallel  *_parallel;
    CPUTime   *_cputime;
    Control   *_control;

    struct timeval et1, et2;
    int       nproc2;

    gettimeofday(&et1,NULL);

    cout << " MEGADOCK ver. "<< VERSION << endl;
    cout << "      megadock@bi.cs.titech.ac.jp   lastupdated: " << LASTUPDATED << endl;
    cout << endl;

#ifdef _OPENMP
    #pragma omp parallel
    {
        nproc2 = omp_get_num_threads();
        if(omp_get_thread_num() == 0) {
            cout << "#Using OpenMP parallelization: " << nproc2 << " threads." << endl;
        }
    }
#else
    nproc2 = 1;
#endif


    _cputime = new CPUTime();
    _cputime->initialize();

    _parallel = new Parallel(nproc2);

    _control = new Control(_cputime,_parallel);
    _control->initialize(argc,argv);
    _control->execute();

    delete _control;
    delete _parallel;

    _cputime->output();

    delete _cputime;

    gettimeofday(&et2,NULL);

    const float elapsed_time = (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
    printf("\n");
    if (elapsed_time < 5.0) {
        printf("Elapsed time                  = %8.3f sec.\n",elapsed_time);
    } else {
        printf("Elapsed time                  = %8.2f sec.\n",elapsed_time);
    }
    return 0;
}
